from __future__ import annotations

from pathlib import Path

import typer

from fmex.app import FMEXApp
from fmex.services import FrameSession, VideoDecodeError

cli = typer.Typer(
    add_completion=False, invoke_without_command=True, no_args_is_help=False
)


def _validate_video_path(video_file: Path) -> None:
    if not video_file.exists():
        raise typer.BadParameter(f"Video file does not exist: {video_file}")
    if not video_file.is_file():
        raise typer.BadParameter(f"Path is not a file: {video_file}")


@cli.callback()
def run(
    video_file: Path = typer.Argument(..., exists=False),
    outdir: Path | None = typer.Option(
        None, "--outdir", help="Directory for PNG exports"
    ),
) -> None:
    """Open a video in the terminal and export selected frames."""
    _validate_video_path(video_file)
    resolved_outdir = outdir if outdir is not None else Path.cwd()

    try:
        session = FrameSession(video_file=video_file, outdir=resolved_outdir)
        app = FMEXApp(session)
        app.run()
    except VideoDecodeError as exc:
        typer.echo(f"Error: {exc}", err=True)
        raise typer.Exit(code=2)


def main() -> None:
    cli()


if __name__ == "__main__":
    main()



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from enum import Enum


class CommandType(str, Enum):
    OPEN = "open"
    NEXT_FRAME = "next_frame"
    PREV_FRAME = "prev_frame"
    SAVE_FRAME = "save_frame"
    QUIT = "quit"


@dataclass(slots=True)
class UserCommand:
    command_type: CommandType
    source: str
    issued_at: datetime
    result: str
    result_reason: str | None = None



from __future__ import annotations

from PIL import Image
from textual import events
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Container, Vertical
from textual.screen import ModalScreen
from textual.widgets import Input
from textual_image.widget import SixelImage

from fmex.models import SaveStatus
from fmex.services import FrameBoundaryError, FrameSession, VideoDecodeError
from fmex.ui.frame_view import StatusLine, build_footer_widgets


class TimeJumpScreen(ModalScreen[str | None]):
    def compose(self) -> ComposeResult:
        with Container(id="time_jump_modal"):
            yield Input(placeholder="Seconds (e.g. 12.5)", id="time_jump_input")

    def on_mount(self) -> None:
        self.styles.align = ("center", "middle")
        self.styles.background = "#020617 70%"
        modal = self.query_one("#time_jump_modal", Container)
        modal.styles.width = 40
        modal.styles.height = 5
        modal.styles.padding = (0, 2)
        modal.styles.border = ("round", "#22d3ee")
        modal.styles.background = "#020617"
        input_widget = self.query_one("#time_jump_input", Input)
        input_widget.focus()

    def on_input_submitted(self, event: Input.Submitted) -> None:
        if event.input.id != "time_jump_input":
            return
        value = event.value.strip()
        self.dismiss(value or None)

    def on_key(self, event: events.Key) -> None:
        if event.key == "escape":
            self.dismiss(None)
            event.prevent_default()


class FMEXApp(App[None]):
    CSS = """
    Screen {
        background: #020617;
        color: #dbeafe;
    }

    #app_root {
        height: 100%;
    }

    #preview_pane {
        height: 1fr;
        width: 100%;
        align: center middle;
    }

    #preview_frame {
        border: round #0ea5e9;
        background: #000814;
        align: center middle;
        width: 100%;
        height: 100%;
    }

    #preview {
        width: auto;
        height: auto;
        max-width: 100%;
        max-height: 100%;
    }

    ControlsLegend {
        background: #0f172a;
        color: #7dd3fc;
        padding: 0 1;
    }

    #status {
        background: #111827;
        color: #e2e8f0;
        padding: 0 1;
    }

    #time_jump_modal {
        border: round #22d3ee;
        background: #020617;
    }

    #time_jump_input {
        background: #020617;
        color: #e2e8f0;
    }
    """

    BINDINGS = [
        Binding("right", "next_frame", "Next"),
        Binding("left", "prev_frame", "Prev"),
        Binding("l", "next_frame", "Next"),
        Binding("h", "prev_frame", "Prev"),
        Binding("shift+right", "step_forward_10", "Next+10"),
        Binding("shift+left", "step_back_10", "Prev-10"),
        Binding("shift+l", "step_forward_10", "Next+10"),
        Binding("shift+h", "step_back_10", "Prev-10"),
        Binding("ctrl+right", "step_forward_100", "Next+100"),
        Binding("ctrl+left", "step_back_100", "Prev-100"),
        Binding("ctrl+l", "step_forward_100", "Next+100"),
        Binding("ctrl+h", "step_back_100", "Prev-100"),
        Binding("j", "focus_time_jump", "Jump"),
        Binding("s", "save_frame", "Save"),
        Binding("q", "quit_app", "Quit"),
    ]

    def __init__(self, session: FrameSession) -> None:
        super().__init__()
        self.session = session
        self.preview = SixelImage(id="preview")
        self.status = StatusLine("Initializing", id="status")

    def compose(self) -> ComposeResult:
        with Vertical(id="app_root"):
            with Container(id="preview_pane"):
                with Container(id="preview_frame"):
                    yield self.preview
            yield from build_footer_widgets()
            yield self.status

    def _set_preview_image(self, image: Image.Image) -> None:
        self.preview.image = image

    def on_mount(self) -> None:
        snap = self.session.get_current_frame()
        self._set_preview_image(snap.image)
        self._update_frame_status(snap)

    def on_resize(self, event: events.Resize) -> None:
        del event
        self.preview.refresh(layout=True)

    def _set_status(self, text: str) -> None:
        self.status.set_status(text)

    def _set_status_with_frame(self, text: str) -> None:
        current = self.session.session.current_frame_index + 1
        total = self.session.total_frames_display()
        self._set_status(f"{text} | Frame {current}/{total}")

    def _update_frame_status(self, snap, message: str | None = None) -> None:
        status = f"Frame {snap.frame_index + 1}/{self.session.total_frames_display()}"
        if message:
            status = f"{message} | {status}"
        self._set_status(status)

    def action_focus_time_jump(self) -> None:
        self.push_screen(TimeJumpScreen(), self._handle_time_jump)
        self._set_status_with_frame("Enter seconds and press Enter")

    def _handle_time_jump(self, value: str | None) -> None:
        if not value:
            self._set_status_with_frame("Jump canceled")
            return
        try:
            seconds = float(value)
            snap, message = self.session.jump_to_time(seconds)
            self._set_preview_image(snap.image)
            self._update_frame_status(snap, message)
        except (ValueError, FrameBoundaryError, VideoDecodeError) as exc:
            self._set_status_with_frame(str(exc))

    def _step_by(self, delta: int) -> None:
        try:
            snap, message = self.session.step_frames(delta)
            self._set_preview_image(snap.image)
            self._update_frame_status(snap, message)
        except FrameBoundaryError as exc:
            self._set_status_with_frame(str(exc))

    def action_step_forward_10(self) -> None:
        self._step_by(10)

    def action_step_back_10(self) -> None:
        self._step_by(-10)

    def action_step_forward_100(self) -> None:
        self._step_by(100)

    def action_step_back_100(self) -> None:
        self._step_by(-100)

    def action_next_frame(self) -> None:
        self._step_by(1)

    def action_prev_frame(self) -> None:
        self._step_by(-1)

    def action_save_frame(self) -> None:
        result = self.session.save_current_frame()
        if result.status == SaveStatus.SUCCESS:
            self._set_status_with_frame(f"Saved: {result.output_path}")
        else:
            self._set_status_with_frame(result.error_message or "Failed to save frame")

    def action_quit_app(self) -> None:
        self.session.close()
        self.exit()



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from uuid import uuid4

from fmex.models import (
    FrameSnapshot,
    SaveOperation,
    SaveStatus,
    SessionStatus,
    VideoSession,
)
from fmex.services.frame_cache import FrameCache
from fmex.services.frame_saver import FrameSaveError, FrameSaver
from fmex.services.video_decoder import (
    FrameIndexError,
    PyAVVideoDecoder,
    VideoDecodeError,
)


class FrameBoundaryError(RuntimeError):
    pass


@dataclass
class FrameSession:
    video_file: Path
    outdir: Path
    decoder_factory: type[PyAVVideoDecoder] = PyAVVideoDecoder

    def __post_init__(self) -> None:
        self.session = VideoSession(
            session_id=uuid4().hex,
            source_path=self.video_file,
            total_frames=0,
            current_frame_index=0,
            outdir=self.outdir,
            status=SessionStatus.INITIALIZING,
        )
        self.cache = FrameCache(max_items=64)
        self.decoder = self.decoder_factory(self.video_file)
        self.saver = FrameSaver(self.outdir)
        self.session.total_frames = self.decoder.frame_count
        self.session.status = SessionStatus.READY

    def _snapshot(self, frame_index: int) -> FrameSnapshot:
        cached = self.cache.get(frame_index)
        image = cached if cached is not None else self.decoder.get_frame(frame_index)
        self.cache.put(frame_index, image)
        return FrameSnapshot(
            session_id=self.session.session_id,
            frame_index=frame_index,
            timestamp_ms=frame_index,
            width=image.width,
            height=image.height,
            image=image,
        )

    def get_current_frame(self) -> FrameSnapshot:
        return self._snapshot(self.session.current_frame_index)

    def _navigate_to_index(
        self,
        target_index: int,
        *,
        low_message: str,
        high_message: str,
        boundary_error: str,
    ) -> tuple[FrameSnapshot, str | None]:
        message = None
        if target_index < 0:
            target_index = 0
            message = low_message
        elif (
            self.session.total_frames > 0 and target_index >= self.session.total_frames
        ):
            target_index = self.session.total_frames - 1
            message = high_message
        try:
            snap = self._snapshot(target_index)
        except FrameIndexError as exc:
            if self.decoder.has_known_frame_count:
                self.session.total_frames = self.decoder.frame_count
            if self.session.total_frames > 0:
                if target_index >= self.session.total_frames:
                    target_index = self.session.total_frames - 1
                    message = high_message
                snap = self._snapshot(target_index)
            else:
                raise FrameBoundaryError(boundary_error) from exc
        self.session.current_frame_index = target_index
        self._prefetch_neighbors()
        return snap, message

    def step_frames(self, delta: int) -> tuple[FrameSnapshot, str | None]:
        if delta == 0:
            return self.get_current_frame(), None
        target_index = self.session.current_frame_index + delta
        return self._navigate_to_index(
            target_index,
            low_message="Already at first frame",
            high_message="Already at last frame",
            boundary_error="Already at last frame",
        )

    def jump_to_time(self, seconds: float) -> tuple[FrameSnapshot, str | None]:
        if seconds < 0:
            raise ValueError("Seconds must be non-negative")
        target_index = self.decoder.frame_index_for_seconds(seconds)
        return self._navigate_to_index(
            target_index,
            low_message="Time out of range; moved to first frame",
            high_message="Time out of range; moved to last frame",
            boundary_error="Unable to jump to time",
        )

    def next_frame(self) -> FrameSnapshot:
        target_index = self.session.current_frame_index + 1
        try:
            snap = self._snapshot(target_index)
        except FrameIndexError as exc:
            if self.decoder.frame_count:
                self.session.total_frames = self.decoder.frame_count
            raise FrameBoundaryError("Already at last frame") from exc
        self.session.current_frame_index = target_index
        self._prefetch_neighbors()
        return snap

    def previous_frame(self) -> FrameSnapshot:
        if self.session.current_frame_index <= 0:
            raise FrameBoundaryError("Already at first frame")
        self.session.current_frame_index -= 1
        snap = self._snapshot(self.session.current_frame_index)
        self._prefetch_neighbors()
        return snap

    def _prefetch_neighbors(self) -> None:
        idx = self.session.current_frame_index
        neighbors = [idx + 1, idx - 1]
        if self.session.total_frames > 0:
            candidates = [n for n in neighbors if 0 <= n < self.session.total_frames]
        else:
            candidates = [n for n in neighbors if n >= 0]
        self.cache.prefetch(
            candidates,
            self.decoder.get_frame,
        )

    def save_current_frame(self) -> SaveOperation:
        now = datetime.now()
        attempted_index = self.session.current_frame_index
        attempted_output = self.saver.build_target(attempted_index)
        try:
            snapshot = self.get_current_frame()
            output = self.saver.save_png(
                snapshot.image,
                snapshot.frame_index,
                target=attempted_output,
            )
            return SaveOperation(
                save_id=uuid4().hex,
                session_id=self.session.session_id,
                frame_index=snapshot.frame_index,
                output_path=output,
                status=SaveStatus.SUCCESS,
                created_at=now,
            )
        except (FrameSaveError, VideoDecodeError) as exc:
            return SaveOperation(
                save_id=uuid4().hex,
                session_id=self.session.session_id,
                frame_index=attempted_index,
                output_path=attempted_output,
                status=SaveStatus.FAILURE,
                created_at=now,
                error_message=str(exc),
            )

    def total_frames_display(self) -> str:
        total = self.session.total_frames
        return str(total) if total > 0 else "?"

    def close(self) -> None:
        self.session.status = SessionStatus.CLOSED
        self.cache.close()
        if hasattr(self.decoder, "close"):
            self.decoder.close()



"""Domain models for fmex."""

from .commands import CommandType, UserCommand
from .session import (
    FrameSnapshot,
    SaveOperation,
    SaveStatus,
    SessionStatus,
    VideoSession,
)

__all__ = [
    "CommandType",
    "FrameSnapshot",
    "SaveOperation",
    "SaveStatus",
    "SessionStatus",
    "UserCommand",
    "VideoSession",
]



from __future__ import annotations

from fmex.models.commands import CommandType

KEY_COMMANDS: dict[str, CommandType] = {
    "right": CommandType.NEXT_FRAME,
    "left": CommandType.PREV_FRAME,
    "l": CommandType.NEXT_FRAME,
    "h": CommandType.PREV_FRAME,
    "s": CommandType.SAVE_FRAME,
    "q": CommandType.QUIT,
}


def command_for_key(key: str) -> CommandType | None:
    return KEY_COMMANDS.get(key)


def controls_text() -> str:
    return (
        "Left/h prev | Right/l next | "
        "Shift+Left/Right ±10 | Ctrl+Left/Right ±100 | "
        "j jump | s Save PNG | q quit"
    )



from __future__ import annotations

from collections import OrderedDict
from io import BytesIO
from pathlib import Path

from PIL import Image

try:
    from diskcache import Cache
except Exception:  # pragma: no cover - optional backend failure
    Cache = None  # type: ignore[misc,assignment]


class FrameCache:
    def __init__(self, max_items: int = 64, disk_dir: Path | None = None) -> None:
        self.max_items = max_items
        self._memory: OrderedDict[int, Image.Image] = OrderedDict()
        self._disk = Cache(str(disk_dir)) if (disk_dir and Cache is not None) else None

    def get(self, frame_index: int) -> Image.Image | None:
        if frame_index in self._memory:
            image = self._memory.pop(frame_index)
            self._memory[frame_index] = image
            return image

        if self._disk is not None:
            payload = self._disk.get(frame_index)
            if payload is None:
                return None
            image = Image.open(BytesIO(payload)).copy()
            self.put(frame_index, image)
            return image

        return None

    def put(self, frame_index: int, image: Image.Image) -> None:
        if frame_index in self._memory:
            self._memory.pop(frame_index)
        self._memory[frame_index] = image
        while len(self._memory) > self.max_items:
            self._memory.popitem(last=False)

        if self._disk is not None:
            buffer = BytesIO()
            image.save(buffer, format="PNG")
            self._disk[frame_index] = buffer.getvalue()

    def prefetch(self, indices: list[int], fetcher: callable) -> None:
        for idx in indices:
            if idx < 0 or self.get(idx) is not None:
                continue
            try:
                image = fetcher(idx)
            except Exception:
                continue
            self.put(idx, image)

    def close(self) -> None:
        if self._disk is not None:
            self._disk.close()



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from pathlib import Path

from PIL import Image


class SessionStatus(str, Enum):
    INITIALIZING = "initializing"
    READY = "ready"
    ERRORED = "errored"
    CLOSED = "closed"


class SaveStatus(str, Enum):
    SUCCESS = "success"
    FAILURE = "failure"


@dataclass(slots=True)
class VideoSession:
    session_id: str
    source_path: Path
    total_frames: int
    current_frame_index: int
    outdir: Path
    status: SessionStatus
    error_message: str | None = None


@dataclass(slots=True)
class FrameSnapshot:
    session_id: str
    frame_index: int
    timestamp_ms: int
    width: int
    height: int
    image: Image.Image


@dataclass(slots=True)
class SaveOperation:
    save_id: str
    session_id: str
    frame_index: int
    output_path: Path
    status: SaveStatus
    created_at: datetime
    error_message: str | None = None



"""Service layer for fmex."""

from .frame_cache import FrameCache
from .frame_saver import FrameSaveError, FrameSaver
from .session_service import FrameBoundaryError, FrameSession
from .video_decoder import FrameIndexError, PyAVVideoDecoder, VideoDecodeError

__all__ = [
    "FrameBoundaryError",
    "FrameCache",
    "FrameSaveError",
    "FrameSaver",
    "FrameSession",
    "FrameIndexError",
    "PyAVVideoDecoder",
    "VideoDecodeError",
]



from __future__ import annotations

from datetime import datetime
from pathlib import Path
from uuid import uuid4

from PIL import Image


class FrameSaveError(RuntimeError):
    pass


class FrameSaver:
    def __init__(self, outdir: Path) -> None:
        self.outdir = outdir

    def ensure_outdir(self) -> Path:
        try:
            self.outdir.mkdir(parents=True, exist_ok=True)
        except Exception as exc:
            raise FrameSaveError(
                f"Failed to create output directory '{self.outdir}': {exc}"
            ) from exc

        if not self.outdir.is_dir():
            raise FrameSaveError(f"Output path is not a directory: {self.outdir}")
        return self.outdir

    def build_filename(self, frame_index: int) -> str:
        stamp = datetime.now().strftime("%Y%m%d-%H%M%S")
        suffix = uuid4().hex[:8]
        return f"frame-{frame_index:06d}-{stamp}-{suffix}.png"

    def build_target(self, frame_index: int) -> Path:
        return self.outdir / self.build_filename(frame_index)

    def save_png(
        self,
        image: Image.Image,
        frame_index: int,
        target: Path | None = None,
    ) -> Path:
        outdir = self.ensure_outdir()
        target = target or (outdir / self.build_filename(frame_index))

        try:
            image.save(target, format="PNG")
        except Exception as exc:
            raise FrameSaveError(f"Failed to save PNG '{target}': {exc}") from exc

        return target



from __future__ import annotations

from pathlib import Path

from PIL import Image

try:
    import av
except Exception:  # pragma: no cover - environment dependent
    av = None  # type: ignore[assignment]


class VideoDecodeError(RuntimeError):
    pass


class FrameIndexError(VideoDecodeError):
    pass


class PyAVVideoDecoder:
    """Lazy decoder optimized for fast startup and on-demand frame access."""

    def __init__(self, video_path: Path, max_cache: int = 256) -> None:
        self.video_path = video_path
        self._max_cache = max_cache
        self._frames: list[Image.Image] = []
        self._cache_start = 0
        self._container = None
        self._stream = None
        self._decode_iter = None
        self._decoded_complete = False
        self._frame_count = 0
        self._fps: float | None = None
        self._open()

    @property
    def frame_count(self) -> int:
        frame_count = getattr(self, "_frame_count", 0)
        if frame_count:
            return frame_count
        frames = getattr(self, "_frames", [])
        cache_start = getattr(self, "_cache_start", 0)
        return cache_start + len(frames)

    @property
    def has_known_frame_count(self) -> bool:
        return bool(getattr(self, "_frame_count", 0))

    @property
    def fps(self) -> float | None:
        return self._fps

    def frame_index_for_seconds(self, seconds: float) -> int:
        if seconds < 0:
            raise ValueError("Seconds must be non-negative")
        if not self._fps:
            raise VideoDecodeError("FPS is unavailable for this video")
        return max(0, int(round(seconds * self._fps)))

    def _open(self) -> None:
        if av is None:
            raise VideoDecodeError("PyAV is not available in this environment")
        if not self.video_path.exists():
            raise VideoDecodeError(f"Video not found: {self.video_path}")

        try:
            self._container = av.open(str(self.video_path))
            self._stream = next(
                (s for s in self._container.streams if s.type == "video"), None
            )
            if self._stream is None:
                raise VideoDecodeError("No video stream found")
            if self._stream.average_rate:
                self._fps = float(self._stream.average_rate)
            self._frame_count = self._infer_frame_count()
            self._decode_iter = self._container.decode(self._stream)
        except VideoDecodeError:
            raise
        except Exception as exc:  # pragma: no cover - depends on media/backend
            raise VideoDecodeError(f"Unable to decode video: {exc}") from exc

    def get_frame(self, frame_index: int) -> Image.Image:
        frames = getattr(self, "_frames", [])
        decode_iter = getattr(self, "_decode_iter", None)
        decoded_complete = getattr(self, "_decoded_complete", False)
        frame_count = getattr(self, "_frame_count", 0)
        cache_start = getattr(self, "_cache_start", 0)
        max_cache = max(1, int(getattr(self, "_max_cache", 1)))
        if frame_index < 0:
            raise FrameIndexError(f"Frame index out of bounds: {frame_index}")
        if frame_count and frame_index >= frame_count:
            raise FrameIndexError(f"Frame index out of bounds: {frame_index}")
        cache_end = cache_start + len(frames)
        if cache_start <= frame_index < cache_end:
            return frames[frame_index - cache_start]
        if frame_index < cache_start:
            if not self._seek_to_frame(frame_index):
                self._reset_decode()
            frames = self._frames
            decode_iter = self._decode_iter
            decoded_complete = self._decoded_complete
            cache_start = self._cache_start
            cache_end = cache_start + len(frames)
            if cache_start <= frame_index < cache_end:
                return frames[frame_index - cache_start]
        if decode_iter is None:
            raise FrameIndexError(f"Frame index out of bounds: {frame_index}")

        if frame_index >= cache_end + max_cache:
            if self._seek_to_frame(frame_index):
                frames = self._frames
                decode_iter = self._decode_iter
                decoded_complete = self._decoded_complete
                cache_start = self._cache_start
                cache_end = cache_start + len(frames)
                if cache_start <= frame_index < cache_end:
                    return frames[frame_index - cache_start]

        while (cache_start + len(frames)) <= frame_index and not decoded_complete:
            try:
                frame = next(decode_iter)
            except StopIteration:
                decoded_complete = True
                self._decoded_complete = True
                break
            except Exception as exc:  # pragma: no cover - backend/media dependent
                raise VideoDecodeError(f"Unable to decode frame: {exc}") from exc
            image = frame.to_image().convert("RGB")
            pts = getattr(frame, "pts", None)
            if not frames and pts is not None and self._fps:
                time_base = getattr(frame, "time_base", None) or self._stream.time_base
                cache_start = self._frame_index_from_pts(pts, time_base)
                self._cache_start = cache_start
            frames.append(image)
            if len(frames) > max_cache:
                frames.pop(0)
                cache_start += 1
                self._cache_start = cache_start
            self._frames = frames

        cache_end = cache_start + len(frames)
        if cache_start <= frame_index < cache_end:
            return frames[frame_index - cache_start]

        if decoded_complete:
            self._frame_count = cache_start + len(frames)
            raise FrameIndexError(f"Frame index out of bounds: {frame_index}")

        raise VideoDecodeError(f"Unable to decode frame: {frame_index}")

    def _frame_index_from_pts(self, pts: int, time_base) -> int:
        if not self._fps or not time_base:
            return self._cache_start
        seconds = float(pts * time_base)
        return max(0, int(round(seconds * self._fps)))

    def _seek_to_frame(self, frame_index: int) -> bool:
        if self._container is None or self._stream is None or not self._fps:
            return False
        time_base = self._stream.time_base
        if not time_base:
            return False
        seconds = frame_index / self._fps
        offset = int(seconds / float(time_base))
        try:
            self._container.seek(
                offset,
                stream=self._stream,
                backward=True,
                any_frame=False,
            )
        except Exception:
            return False
        self._frames = []
        self._cache_start = 0
        self._decoded_complete = False
        self._decode_iter = self._container.decode(self._stream)
        return True

    def _infer_frame_count(self) -> int:
        if self._stream is None or self._container is None:
            return 0
        if self._stream.frames:
            return int(self._stream.frames)
        if self._stream.duration and self._stream.average_rate:
            seconds = float(self._stream.duration * self._stream.time_base)
            fps = float(self._stream.average_rate)
            if seconds > 0 and fps > 0:
                return max(1, int(round(seconds * fps)))
        if self._container.duration and self._stream.average_rate:
            seconds = float(self._container.duration) / 1_000_000.0
            fps = float(self._stream.average_rate)
            if seconds > 0 and fps > 0:
                return max(1, int(round(seconds * fps)))
        return 0

    def _reset_decode(self) -> None:
        if self._container is not None:
            self._container.close()
        self._frames = []
        self._cache_start = 0
        self._decoded_complete = False
        self._decode_iter = None
        self._container = None
        self._stream = None
        self._open()

    def close(self) -> None:
        if self._container is not None:
            self._container.close()
            self._container = None
            self._stream = None
            self._decode_iter = None



from __future__ import annotations

import re

from rich.text import Text
from textual.widget import Widget
from textual.widgets import Static

from fmex.ui.keymap import controls_text


class ControlsLegend(Static):
    def __init__(self) -> None:
        self.text = controls_text()
        super().__init__(self.text)


class StatusLine(Static):
    _FRAME_COUNTER_PATTERN = re.compile(r"Frame \d+/(?:\d+|\?)")

    def __init__(self, text: str, **kwargs) -> None:
        self.text = text
        super().__init__(self._render_status(text), **kwargs)

    def set_status(self, text: str) -> None:
        self.text = text
        self.update(self._render_status(text))

    def _render_status(self, text: str) -> Text:
        rendered = Text(text)
        for match in self._FRAME_COUNTER_PATTERN.finditer(text):
            rendered.stylize("bold #f8fafc on #0369a1", match.start(), match.end())
        return rendered


def build_footer_widgets() -> list[Widget]:
    return [ControlsLegend()]
